from abc import ABC, abstractmethod
from typing import Any, Dict, Tuple


class AbstractAsyncSession(ABC):
    started: bool = False  # Changes according to implementation

    @abstractmethod
    async def start(self, instructions: Dict[str, Tuple[Tuple, Dict[str, Any]]]):
        """
        :param instructions:
        Settings from TelegramClient. Example:
            {
                <function> SessionMethod: <tuple[2]> (
                    <tuple...> (...*_params_args),
                    <dict> {...*_params_kwargs}
                )
            }
        """
        raise NotImplementedError

    @abstractmethod
    async def set_dc(self, dc_id, server_address, port):
        """
        Sets the information of the data center address and port that
        the library should connect to, as well as the data center ID,
        which is currently unused.
        """
        raise NotImplementedError

    @abstractmethod
    async def dc_id(self):
        """
        Returns the currently-used data center ID.
        """
        raise NotImplementedError

    @abstractmethod
    async def server_address(self):
        """
        Returns the server address where the library should connect to.
        """
        raise NotImplementedError

    @abstractmethod
    async def port(self):
        """
        Returns the port to which the library should connect to.
        """
        raise NotImplementedError

    @abstractmethod
    async def auth_key(self):
        """
        Returns an ``AuthKey`` instance associated with the saved
        data center, or `None` if a new one should be generated.
        """
        raise NotImplementedError

    @abstractmethod
    async def set_auth_key(self, auth_key: str):
        """
        Sets the ``AuthKey`` to be used for the saved data center.
        """
        raise NotImplementedError

    @abstractmethod
    async def takeout_id(self):
        """
        Returns an ID of the takeout process initialized for this session,
        or `None` if there's no were any unfinished takeout requests.
        """
        raise NotImplementedError

    @abstractmethod
    async def set_takeout_id(self, takeout_id: int):
        """
        Sets the ID of the unfinished takeout process for this session.
        """
        raise NotImplementedError

    @abstractmethod
    async def get_update_state(self, entity_id):
        """
        Returns the ``UpdateState`` associated with the given `entity_id`.
        If the `entity_id` is 0, it should return the ``UpdateState`` for
        no specific channel (the "general" state). If no state is known
        it should ``return None``.
        """
        raise NotImplementedError

    @abstractmethod
    async def set_update_state(self, entity_id, state):
        """
        Sets the given ``UpdateState`` for the specified `entity_id`, which
        should be 0 if the ``UpdateState`` is the "general" state (and not
        for any specific channel).
        """
        raise NotImplementedError

    @abstractmethod
    async def close(self):
        """
        Called on client disconnection. Should be used to
        free any used resources. Can be left empty if none.
        """

    @abstractmethod
    async def save(self):
        """
        Called whenever important properties change. It should
        make persist the relevant session information to disk.
        """
        raise NotImplementedError

    @abstractmethod
    async def delete(self):
        """
        Called upon client.log_out(). Should delete the stored
        information from disk since it's not valid anymore.
        """
        raise NotImplementedError

    @classmethod
    async def list_sessions(cls):
        """
        Lists available sessions. Not used by the library itself.
        """
        return []

    @abstractmethod
    async def process_entities(self, tlo):
        """
        Processes the input ``TLObject`` or ``list`` and saves
        whatever information is relevant (e.g., ID or access hash).
        """
        raise NotImplementedError

    @abstractmethod
    async def get_input_entity(self, key):
        """
        Turns the given key into an ``InputPeer`` (e.g. ``InputPeerUser``).
        The library uses this method whenever an ``InputPeer`` is needed
        to suit several purposes (e.g. user only provided its ID or wishes
        to use a cached username to avoid extra RPC).
        """
        raise NotImplementedError

    @abstractmethod
    async def cache_file(self, md5_digest, file_size, instance):
        """
        Caches the given file information persistently, so that it
        doesn't need to be re-uploaded in case the file is used again.

        The ``instance`` will be either an ``InputPhoto`` or ``InputDocument``,
        both with an ``.id`` and ``.access_hash`` attributes.
        """
        raise NotImplementedError

    @abstractmethod
    async def get_file(self, md5_digest, file_size, cls):
        """
        Returns an instance of ``cls`` if the ``md5_digest`` and ``file_size``
        match an existing saved record. The class will either be an
        ``InputPhoto`` or ``InputDocument``, both with two parameters
        ``id`` and ``access_hash`` in that order.
        """
        raise NotImplementedError



"""
---
mod stands for module not modulo :)
"""

import inspect
import importlib
from typing import TypeVar, Tuple, cast

AMod = TypeVar("AMod")
BMod = TypeVar("BMod")

IS_PATCHED = False


def do_import(
    a_mod: str,
    klass: str,
    b_mod_replace: Tuple[str, str]
) -> Tuple[AMod, BMod]:
    a_lib_impl = importlib.import_module(a_mod)
    b_lib_impl = importlib.import_module(a_mod.replace(*b_mod_replace))

    return (
        cast(AMod, getattr(a_lib_impl, klass)),
        cast(BMod, getattr(b_lib_impl, klass))
    )


def patch(a_type: AMod, b_type: BMod, *include: str) -> None:
    for method in include:
        if method == "*":
            for name, b_obj in inspect.getmembers(b_type):
                setattr(a_type, name, b_obj)

            return
        setattr(a_type, method, getattr(b_type, method))


try:
    import telethon

    B_REPLACE = ("telethon", "telethon_asyncpg")

    A, B = do_import("telethon.client.auth", "AuthMethods", B_REPLACE)
    patch(A, B, "_start", "start")

    A, B = do_import("telethon.client.downloads", "_DirectDownloadIter", B_REPLACE)
    patch(A, B, "_init")

    A, B = do_import("telethon.client.telegrambaseclient", "TelegramBaseClient", B_REPLACE)
    patch(A, B, "__init__", "connect", "_disconnect", "_switch_dc", "_auth_key_callback")

    A, B = do_import("telethon.client.updates", "UpdateMethods", B_REPLACE)
    patch(A, B, "_handle_update", "_update_loop", "_dispatch_update")

    A, B = do_import("telethon.client.users", "UserMethods", B_REPLACE)
    patch(A, B, "_call", "get_input_entity", )

    A, B = do_import("telethon.network.mtprotosender", "MTProtoSender", B_REPLACE)
    patch(A, B, "_try_gen_auth_key", )

    IS_PATCHED = True
except ImportError as exc:
    raise RuntimeError("Remember: `patch` will not work without telethon") from exc


__all__ = []



"""
todo(refactor): make more abstract BaseAsyncSession
Session shouldn't know anything about tl/* even session's private interface
"""

from abc import ABC

from ..sessions.abstract import AbstractAsyncSession
from telethon import utils
from telethon.tl import types


class BaseAsyncSession(AbstractAsyncSession, ABC):
    def __init__(self):
        self._dc_id = 0
        self._server_address = None
        self._port = None
        self._auth_key = None
        self._takeout_id = None

        self._files = {}
        self._entities = set()
        self._update_states = {}

    @property
    def dc_id(self):
        return self._dc_id

    @property
    def server_address(self):
        return self._server_address

    @property
    def port(self):
        return self._port

    @property
    def auth_key(self):
        return self._auth_key

    @property
    def takeout_id(self):
        return self._takeout_id

    @classmethod
    def _entity_values_to_row(cls, id, hash, username, phone, name):
        # While this is a simple implementation it might be overrode by,
        # other classes so they don't need to implement the plural form
        # of the method. Don't remove.
        return id, hash, username, phone, name

    def _entity_to_row(self, e):
        if not isinstance(e, types.TLObject):
            return
        try:
            p = utils.get_input_peer(e, allow_self=False)
            marked_id = utils.get_peer_id(p)
        except TypeError:
            # Note: `get_input_peer` already checks for non-zero `access_hash`.
            #        See issues #354 and #392. It also checks that the entity
            #        is not `min`, because its `access_hash` cannot be used
            #        anywhere (since layer 102, there are two access hashes).
            return

        if isinstance(p, (types.InputPeerUser, types.InputPeerChannel)):
            p_hash = p.access_hash
        elif isinstance(p, types.InputPeerChat):
            p_hash = 0
        else:
            return

        username = getattr(e, 'username', None) or None
        if username is not None:
            username = username.lower()
        phone = getattr(e, 'phone', None)
        name = utils.get_display_name(e) or None
        return self._entity_values_to_row(
            marked_id, p_hash, username, phone, name
        )

    def _entities_to_rows(self, tlo):
        if not isinstance(tlo, types.TLObject) and utils.is_list_like(tlo):
            # This may be a list of users already for instance
            entities = tlo
        else:
            entities = []
            if hasattr(tlo, 'user'):
                entities.append(tlo.user)
            if hasattr(tlo, 'chats') and utils.is_list_like(tlo.chats):
                entities.extend(tlo.chats)
            if hasattr(tlo, 'users') and utils.is_list_like(tlo.users):
                entities.extend(tlo.users)

        rows = []  # Rows to add (id, hash, username, phone, name)
        for e in entities:
            row = self._entity_to_row(e)
            if row:
                rows.append(row)
        return rows




"""
schema for asyncpg sessions is "asyncpg_telethon"
"""

from typing import List, Optional, Dict, Any, Union, Callable
import asyncio
import uuid
from abc import ABC
import datetime
import logging

import asyncpg

from telethon import utils
from telethon.crypto import AuthKey
from telethon.tl import types
from ..sessions.base import BaseAsyncSession

TABLES = ("sessions", "sent_files", "update_state",)
TELETHON_SQLITE_CURRENT_VERSION = 6  # database versions must be the same as telethon's original SQLite version
ALLOWED_ENTITY_IDENTIFIER_NAMES = ("name", "username", "phone", )

logger = logging.getLogger(__name__)

_sfconf = {
    types.InputDocument: 0,
    types.InputPhoto: 1,
}

_sfconf_keys = tuple(_sfconf.keys())


def default_session_id_factory() -> str:
    return str(uuid.uuid4())


def _sftype(cls):
    if cls not in _sfconf:
        raise ValueError('The cls must be either InputDocument/InputPhoto')
    return _sfconf[cls]


async def check_tables(connection: asyncpg.Connection) -> bool:
    for table in TABLES:
        rec = await connection.fetchval(
            """
            select EXISTS(
            select 1
            from pg_tables
            where schemaname = $1 and tablename = $2);
            """, "asyncpg_telethon", table
        )

        if bool(rec) is False:
            logger.debug(f"Table = {table} with schema `asyncpg_telethon` does not exist")
            return False

    return True


async def create_tables(connection: asyncpg.Connection, lock: asyncio.Lock):
    create_tables_sql = (
        """sessions (
            session_id varchar(255),
            dc_id integer,
            server_address text,
            port integer,
            auth_key bytea,
            takeout_id integer,
            primary key(session_id, dc_id)
        )""",
        """entities (
            session_id varchar(255),
            id bigint,
            hash bigint not null,
            username text ,
            phone bigint default null,
            name text,
            primary key(session_id, id)
        )""",
        """sent_files (
            session_id varchar(255),
            md5_digest bytea,
            file_size integer,
            type integer,
            id bigint,
            hash bigint,
            primary key(session_id, md5_digest, file_size, type)
        )""",
        """update_state (
            session_id varchar(255),
            id integer,
            pts integer,
            qts integer,
            date integer,
            seq integer,
            primary key(session_id, id)
        )""")

    async with lock:
        logger.debug(f"Creating schema(`asyncpg_telethon`) and tables: {TABLES} with schema `asyncpg_telethon`")
        async with connection.transaction(isolation="read_committed"):
            await connection.execute("""create schema if not exists "asyncpg_telethon";""")
            # table is sure safe to be passed by f'' to query.
            await connection.execute("".join(
                f"""create table if not exists "asyncpg_telethon".{table};"""
                for table in create_tables_sql
            ))
        logger.debug("Tables created")


class AsyncpgSession(BaseAsyncSession, ABC):
    def __init__(
        self,
        asyncpg_conf: Optional[Union[str, Dict[str, Any]]],
        session_id_factory: Callable[[], str] = default_session_id_factory,
        save_entities: bool = True
    ):
        """
        Initializer for AsyncpgSession
        :param asyncpg_conf: asyncpg.create_pool configs
        :param session_id_factory: function to generate session identifier
        :param save_entities: True - all entities will be cached while processing
        """
        super().__init__()
        self._session_id = session_id_factory()
        self.save_entities = save_entities
        self._dsn = asyncpg_conf
        self._pool: Optional[asyncpg.pool.Pool] = None
        self._lock = asyncio.Lock()
        self._conn: Optional[asyncpg.Connection] = None

    @classmethod
    def with_pool(
        cls,
        asyncpg_pool: asyncpg.pool.Pool,
        session_id_factory: Callable[[], str] = default_session_id_factory,
        save_entities: bool = True
    ) -> 'AsyncpgSession':
        """
        Another implementation of initializer but for shared pool
        :param asyncpg_pool: ready asyncpg Pool
        :param session_id_factory: ...
        :param save_entities: ...
        :return: instance of AsyncpgSession
        """
        self = cls(
            asyncpg_conf=None,
            session_id_factory=session_id_factory,
            save_entities=save_entities
        )  # type: ignore
        self._pool = asyncpg_pool
        return self

    async def start(self, settings: dict):
        if self.started:
            return

        if not isinstance(self._pool, asyncpg.pool.Pool):
            # no more than one pool should be opened in one instance
            self._pool = await asyncpg.create_pool(**self._dsn)

        async with self._pool.acquire() as conn:  # type: asyncpg.Connection
            if (await check_tables(conn)) is False:
                await create_tables(conn, self._lock)

            async with self._lock:
                for method, (args, kwargs) in settings.items():
                    try:
                        self._conn = conn
                        await method(*args, **kwargs)
                    except asyncpg.InterfaceError as exc:
                        await self.close()
                        raise exc
                else:
                    self.started = True
                    self._conn = None
                    await conn.close()

    async def set_dc(self, dc_id, server_address, port):
        # Fetch the auth_key corresponding to this data center
        self._dc_id = dc_id
        self._port = port
        self._server_address = server_address
        await self._update_session_table()

    async def set_auth_key(self, auth_key: AuthKey):
        self._auth_key = AuthKey(data=auth_key)
        await self._update_session_table()

    async def set_takeout_id(self, takeout_id: int):
        self._takeout_id = takeout_id
        await self._update_session_table()

    async def _update_session_table(self):
        """
        Method is not responsible to close connection he hasn't opened.
        """
        query, args = """
            insert into 
            asyncpg_telethon.sessions (session_id, dc_id, server_address, port, auth_key, takeout_id) 
            values ($1, $2, $3, $4, $5, $6) 
            on conflict(session_id, dc_id) do 
            update set dc_id = $2,
            server_address = $3, port = $4,
            auth_key = $5, takeout_id = $6 
            where sessions.session_id = $1;
        """, (
            self._session_id,
            self._dc_id,
            self._server_address,
            self._port,
            self._auth_key.key if self._auth_key else b'',
            self._takeout_id
        )

        if self._conn is not None:
            await self._conn.execute(query, *args)
            return

        else:
            async with self._pool.acquire() as conn:  # type: asyncpg.Connection
                await conn.execute(query, *args)

    async def get_update_state(self, entity_id):
        query = """
            select pts, qts, date, seq
            from asyncpg_telethon.update_state 
            where update_state.session_id = $1 and id = $2;
        """

        async with self._pool.acquire() as conn:  # type: asyncpg.Connection
            row = await conn.fetchrow(
                query, self._session_id, entity_id
            )
            if row:
                pts, qts, date, seq = row.values()
                date = datetime.datetime.fromtimestamp(date, tz=datetime.timezone.utc)
                return types.updates.State(pts, qts, date, seq, unread_count=0)

    async def set_update_state(self, entity_id, state, connection: asyncpg.Connection = None):
        query = """
            insert into asyncpg_telethon.update_state(session_id, id, pts, qts, date, seq) values ($2,$3,$4,$5,$6,$7) 
            on conflict(session_id, id) do 
            update set id = $2, pts = $3, qts = $4, date = $5, seq = $6 
            where update_state.session_id = $1;
        """
        async with self._pool.acquire() as conn:  # type: asyncpg.Connection
            await conn.execute(
                query, self._session_id, entity_id, state.pts,
                state.qts, state.date.timestamp(), state.seq
            )

    async def process_entities(self, tlo):
        """Processes all the found entities on the given TLObject,
           unless .enabled is False.

           Returns True if new input entities were added.
        """
        if not self.save_entities:
            return

        rows = self._entities_to_rows(tlo)
        if not rows:
            return

        query = """
            insert into asyncpg_telethon.entities(session_id, id, hash, username, phone, name) 
            values ($1,$2,$3,$4,$5,$6) 
            on conflict(session_id, id) do 
            update set id = $2, hash = $3, username = $4, phone = $5, name = $6 
            where entities.session_id = $1;
        """

        for n, row in enumerate(rows):
            row = list(row)
            row.insert(0, self._session_id)
            rows[n] = row

        async with self._pool.acquire() as conn:  # type: asyncpg.Connection
            await conn.executemany(query, rows)

    async def _get_entities_by_x(self, coln: str, colval: str) -> List[asyncpg.Record]:
        """
        _get_entities_by_x should never been called from outside.
        """
        if coln not in ALLOWED_ENTITY_IDENTIFIER_NAMES:
            raise RuntimeWarning(f"{coln!s} is not a valid tablename for entity")

        query = """
            select id, hash from asyncpg_telethon.entities
            where 
            entities.session_id = $1 and $2 = $3;
        """

        async with self._pool.acquire() as conn:  # type: asyncpg.Connection
            return await conn.fetch(query, self._session_id, coln, colval,)

    async def get_entity_rows_by_phone(self, phone):
        return await self._get_entities_by_x("phone", phone)

    async def get_entity_rows_by_username(self, username):
        return await self._get_entities_by_x("username", username)

    async def get_entity_rows_by_name(self, name):
        return await self._get_entities_by_x("name", name)

    async def get_entity_rows_by_id(self, id, exact=True):
        async with self._pool.acquire() as conn:  # type: asyncpg.Connection
            if exact:
                return await conn.fetch(
                    """
                    select id, hash from asyncpg_telethon.entities
                    where
                    entities.session_id = $1 and id = $2
                    """,
                    self._session_id,
                    id
                )

            return await conn.fetch(
                """
                select id, hash from asyncpg_telethon.entities
                where
                entities.session_id = $1 and id in ($2,$3,$4)
                """,
                self._session_id,
                utils.get_peer_id(types.PeerUser(id)),
                utils.get_peer_id(types.PeerChat(id)),
                utils.get_peer_id(types.PeerChannel(id))
            )

    # File processing

    async def get_file(self, md5_digest, file_size, cls):
        query = """
            select id, hash from asyncpg_telethon.sent_files 
            where
                sent_files.session_id = $1 
                and 
                md5_digest = $2
                and
                file_size = $3
                and
                type = $4;
        """
        async with self._pool.acquire() as conn:  # type: asyncpg.Connection
            row = await conn.fetchrow(
                query,
                self._session_id,
                md5_digest, file_size, _sftype(cls)
            )
            if row:
                # Both allowed classes have (id, access_hash) as parameters
                return cls(row.get("id"), row.get("hash"))

    async def cache_file(self, md5_digest, file_size, instance):
        if not isinstance(instance, _sfconf_keys):
            raise TypeError('Cannot cache %s instance' % type(instance))

        query = """
            insert into asyncpg_telethon.sent_files(session_id,md5_digest,file_size,type,id,hash) 
            values ($1,$2,$3,$4,$5,$6) 
            on conflict(session_id, md5_digest, file_size, type) do 
            update set md5_digest = $2, file_size = $3, type = $4, id = $5, hash = $6 
            where sent_files.session_id = $1;
        """

        async with self._pool.acquire() as conn:  # type: asyncpg.Connection
            await conn.execute(
                query,
                self._session_id,
                md5_digest, file_size,
                _sftype(type(instance)),
                instance.id,
                instance.access_hash
            )

    async def delete(self):
        async with self._pool.acquire() as conn:
            async with conn.transaction():
                for table in TABLES:
                    await conn.execute(
                        """
                        delete from asyncpg_telethon.$1
                        where
                        $1.session_id = $2;
                        """,
                        table, self._session_id
                    )

    async def get_input_entity(self, key):
        try:
            if key.SUBCLASS_OF_ID in (0xc91c90b6, 0xe669bf46, 0x40f202fd):
                # hex(crc32(b'InputPeer', b'InputUser' and b'InputChannel'))
                # We already have an Input version, so nothing else required
                return key
            # Try to early return if this key can be casted as input peer
            return utils.get_input_peer(key)
        except (AttributeError, TypeError):
            # Not a TLObject or can't be cast into InputPeer
            if isinstance(key, types.TLObject):
                key = utils.get_peer_id(key)
                exact = True
            else:
                exact = not isinstance(key, int) or key < 0

        result = None
        if isinstance(key, str):
            phone = utils.parse_phone(key)
            if phone:
                result = await self.get_entity_rows_by_phone(phone)
            else:
                username, invite = utils.parse_username(key)
                if username and not invite:
                    result = await self.get_entity_rows_by_username(username)
                else:
                    tup = utils.resolve_invite_link(key)[1]
                    if tup:
                        result = await self.get_entity_rows_by_id(tup, exact=False)

        elif isinstance(key, int):
            result = await self.get_entity_rows_by_id(key, exact)

        if not result and isinstance(key, str):
            result = await self.get_entity_rows_by_name(key)

        if result:
            entity_id, entity_hash = result  # unpack resulting tuple
            entity_id, kind = utils.resolve_id(entity_id)
            # removes the mark and returns type of entity
            if kind == types.PeerUser:
                return types.InputPeerUser(entity_id, entity_hash)
            elif kind == types.PeerChat:
                return types.InputPeerChat(entity_id)
            elif kind == types.PeerChannel:
                return types.InputPeerChannel(entity_id, entity_hash)
        else:
            raise ValueError('Could not find input entity with key ', key)

    async def close(self, timeout: int = None):
        """
        Implements connection pool closing.

        :param timeout: Ignored in this implementation
        """
        await self._pool.close()
        self.started = False

    async def save(self):
        """
        AsyncpgSession does not define save metthod.
        """



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['telethon_asyncpg',
 'telethon_asyncpg.client',
 'telethon_asyncpg.network',
 'telethon_asyncpg.sessions']

package_data = \
{'': ['*']}

install_requires = \
['Telethon>=1.13.0,<2.0.0', 'asyncpg>=0.20.1,<0.21.0']

setup_kwargs = {
    'name': 'telethon-asyncpg',
    'version': '0.1.0',
    'description': 'Asyncpg powered session for pgsql databases',
    'long_description': 'Async session for telethon\n==========================\n\nInstallation\n============\n\n::\n\n    pip install telethon_asyncpg\n    # Or with poetry\n    poetry add telethon_asyncpg\n\n\nUsage\n=====\n\n.. code-block:: python3\n\n    import ssl  # optional\n\n    from telethon import events, TelegramClient\n\n    from telethon_asyncpg import AsyncpgSession, install\n    install()\n\n    URI = ???  # URI-string\n    # dialect+driver://username:password@host:port/database\n\n    pgconf = dict(dsn=URI, min_size=5, max_size=5)\n    # to overcome problem with TLS connection to db pass\n    # ssl=ssl.SSLContext(protocol=ssl.PROTOCOL_TLS) to pgconf\n    session = AsyncpgSession(pgconf, session_id_factory=???)\n    # session_id_factory is any callable with "() -> str" signature\n    # default factory is uuid4 str generator. why factory? (it\'s not really factory ik)\n    bot = TelegramClient(session=session, api_id=???, api_hash=???)\n\n    @bot.on(events.NewMessage())\n    async def message_handler(message):\n        await message.reply("Hi!")\n\n    async def start():\n        await bot.start()\n        print(await bot.get_entity("martin_winks"))\n        await bot.run_until_disconnected()\n\n    if __name__ == \'__main__\':\n        import asyncio\n        asyncio.get_event_loop().run_until_complete(start())\n\n- `AsyncpgSession` can also use shared pool by `AsyncpgSession.with_pool` initializer-method\n\n.. code-block:: python\n\n    my_pool = asyncpg.create_pool(...)\n    session = AsyncpgSession.with_pool(my_pool, lambda: "session-id", True)\n\n\nCheck out the ``examples/`` folder for more realistic examples.\n\nContribution\n============\n\nCurrently we have only asyncpg session available, if you want to contribute with your wrapper - welcome. Take `AsyncpgSession` as an example.\n\nFor contributors\n================\n\nPatched TelegramClient <-> Session\n-----------------------------------\n\n- `TelegramClient` and `Session` object share `settings` `{session.meth: (args_seq, kwargs_mapping)}` dictionary. By protocol `TelegramClient` must add callable with args and kwargs. By protocol session must call this functions at start as it wants (e.g. pass more arguments such as `connection` object to session.method)\n\n- `TelegramClient` may call `session.start` several times per one session instance and session should control its start itself and if it\'s already started it shouldn\'t start again\n\n- `Session.save` method is guaranteed to be called as in usual telethon\n\n\nHacking\n-------\n\n::\n\n    # install poetry dependency manager\n    # Fork/Fork+Clone && cd {{cloned}}\n    poetry install\n    # happy hacking!\n\n\nReferences\n==========\n\nTelethon: `here <https://github.com/LonamiWebs/telethon>`_\nasyncpg pg-driver: `asyncpg <https://github.com/MagicStack/asyncpg>`_\n',
    'author': 'mpa',
    'author_email': 'mpa@snejugal.ru',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/ukinti/telethon_asyncpg',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
